DOMAIN = "dictybase"
# FILENAME = "information_mappings.pck"
# TAGS = ["Dictyostelium discoideum", "gene", "essential", "dictyBase"]

PHENOTYPES_FILENAME = "mutants.pkl"
PHENOTYPES_TAGS = ["Dictyostelium discoideum", "mutant", "dictyBase", "phenotype"]
PHENOTYPES_TITLE = 'dictyBase mutant phenotypes'

DICTY_INFO_URL = 'http://www.dictybase.org/db/cgi-bin/dictyBase/download/download.pl?area=general&ID=gene_information.txt'
DICTY_MAPPING_URL = 'http://www.dictybase.org/db/cgi-bin/dictyBase/download/download.pl?area=general&ID=DDB-GeneID-UniProt.txt'